/************************************************
 * SERexxObex.cmd   (c) Franz Bakan  2007 01 09 *
 ************************************************

   REXX script for listing, copying and deleting files from mobile phones.
   Initial version (0.1) written January 2007

   Works at least with a Sony Ericsson K310i connected via irda to a IBM T23
   running OS/2 aka eComStation with Object REXX.

   Might work for other mobile phones with or without modifications.

             !!!!!!!!!!!!!!!!!!!!!!!!
   Warning:  ! Use at your own risk !
             !!!!!!!!!!!!!!!!!!!!!!!!

   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   ! I only tested this script with my SE K310i !
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

   Send me your feedback.
   Contact information: fbakan@gmx.net   www.fbakan.de

   Needs the DLLs from:
    - rxasyn20.zip   http://hobbes.nmsu.edu/pub/os2/dev/rexx/rxasyn20.zip
    - rexxuni20.zip  http://hobbes.nmsu.edu/pub/os2/dev/rexx/rexxuni20.zip

   Might work on other Operating Systems if you replace the fuctionality
   of these 2 DLLs (Async I/O and Unicode-Conversions) with something
   available for your OS. 

   Thanks to
   ------
    Jane ( jane.geekheim.de )
    Jan Kiszka ( www.ircomm2k.de )
    Kolo ( http://www.project2.mhost.ru/list.php?2 )
    Fredrik ( http://openobex.sourceforge.net )
    Alek Winner

   Usage: - Edit the costumizing area to fit your needs,
   -----  - make shure you have a irda (may be serial) connction between
            your mobile phone and your computer
          - then run this script.
          - You should see a listing of the content of your mobile
            (files and folders) Example:

             0 RootDir [TopDir]
             1 Bilder [SubDir]       2 Sounds [SubDir]       3 Designs [SubDir]
             4 Videos [SubDir]       5 Andere [SubDir]

          - Now you can enter number 1 to change to folder 'Bilder'
            If some folder contains files you have something like:

             0 RootDir [TopDir]
             1 DSC00009.JPG 27.8 kB  2 DSC00010.JPG 41.0 kB  3 DSC00014.JPG 28.6 kB
             4 DSC00020.JPG 39.6 kB

          - Now you can enter
             0   to change to the root directory,
             g2  to copy the file 'DSC00010.JPG' to your disk
             g*  to copy all 4 files to your disk
             d4  to delete 'DSC00020.JPG' from the mobile phone
             d*  to delete all files in this folder from the mobile phone
             q   to quit the program
*/

tab='09'x
crlf='0D 0A'x

/* begin of user customizing area */

port="COM1" /* set to the comport you use for irda (or serial connection)  */
col.numatrib='1B'x||'[1m'      /* color for the numbers [1m means 'bold' (ansi) */
col.foldatrib='1B'x||'[36;40m' /* color for the folders [36;40m = cyan on black */
col.fileatrib='1B'x||'[33;40m' /* color for the files [36;40m = yellow on black */
col.defatrib='1B'x||'[0m'      /* default color */
columns=100 /* size of */
rows=50     /* window  */
listsep=tab /* separator for listings, example: tab or crlf  */
debug=0     /* If it does not work, set to 5 to get most verbose debugfile output  */ 

/* end of user customizing area */

mode columns rows

Call RxFuncAdd 'SysCurPos', 'RexxUtil', 'SysCurPos'
Call RxFuncAdd 'UniLoadFuncs', 'RexxUni', 'UniLoadFuncs'
Call UniLoadFuncs
call RxFuncAdd 'RxAsyncLoadFuncs', 'RXASYNC', 'RxAsyncLoadFuncs'
call RxAsyncLoadFuncs

PortHandle=0
Remaining=0

logfile=date('S')||left(time(),2)||substr(time(),4,2)||'.log'
buflen='FF DC'x /* (65500) OBEX_MAX_PACKET_LENGTH largest packet for us */
cp=UniQueryCp() /* query current codepage */
if debug > 0 then call lineout logfile, logfile||'SERexxObex Logfile'||crlf

/* some OBEX constants */
O_CMD.Connect="80"x
O_CMD.Disconnect='81'x
O_CMD.Put="82"x
O_CMD.Get="83"x
O_GetFolderListing_Packet='83 00 1C 42 00 19'x||'x-obex/folder-listing'||'00'x
O_CMD.SetPath="85"x
O_CMD.Abort="FF"x

O_VERSION="10"x
O_FLAGS="00"x

O_HI.Name="01"x
O_HI.Target="46"x
O_HI.Cont='90'x
O_HI.Sucess='A0'x

/* just some UUID. Any 128bit value shoud work */
O_UUID="F9 EC 7B C4 95 3C 11 D2 98 4E 52 54 00 DC 9E 09"x

/* building the obex connect packet */
O_CONNECT_OH.1.HI=O_HI.Target
O_CONNECT_OH.1.content=O_UUID
O_CONNECT_OH.1.len=length(O_CONNECT_OH.1.content)+3
O_ConnectTotalLen=7+O_CONNECT_OH.1.len
O_Connect_Complete=O_CMD.Connect||x2c(d2x(O_ConnectTotalLen,4))||O_VERSION||O_FLAGS||buflen||O_CONNECT_OH.1.HI||x2c(d2x(O_CONNECT_OH.1.len,4))||O_CONNECT_OH.1.content

/* parts of the SetPath packet */
O_SetPath_OH.1.HI=O_HI.Name
O_SetPath_OH.1.flags='02'x
O_SetPath_OH.1.constants='00'x

if debug > 1 then say "open irda..."
/* opening serial communication */
call irda_open
/* setting the mobile in OBEX-mode */
call irda_write_at 0 "AT*EOBEX"
/* do the obex connect */
call irda_write 0 O_Connect_Complete
buffer = irda_read_obex_packet()
if left(buffer,1) <> O_HI.Sucess then say 'We have a problem (1), rc= ' c2x(left(buffer,1))
connection_id=substr(buffer,9,4)
if debug > 3 then  say c2x(connection_id)

do forever /* main loop for user input */
 call listcontent
 if listsep <> crlf then say
 say " Your choice: ["col.numatrib||"n"||col.defatrib||"|"||col.numatrib||"Gn"||col.defatrib||"|"||col.numatrib||"Dn"||col.defatrib||"|"||col.numatrib||"Q"||col.defatrib||"]"
 say " Example: "||col.numatrib||"2"||col.defatrib||" change folder, "||col.numatrib||"g13"||col.defatrib||" get file 13, "||col.numatrib||"d*"||col.defatrib||" delete all files, "||col.numatrib||"q"||col.defatrib||" quit"
 pull object
 select
  when object = 'Q' then leave  /* Quit */
  when datatype(object,'W') = 1 then do
   if object <= folders.0 then call setpath object
   end
  when left(object,1) = 'D' then do /* delete */
   num = substr(object,2)
   select
    when num = '*' then  do i= 1 to files.name.0
     call delfile i
     end
    when datatype(num,'W') = 1 then do
     if num > 0 & num <= files.name.0 then call delfile num
     end
    otherwise
   end
  end
  when left(object,1) = 'G' then do /* get */
   num = substr(object,2)
   select
    when num = '*' then  do i= 1 to files.name.0
     call getfile i
     end
    when datatype(num,'W') = 1 then do
     if num > 0 & num <= files.name.0 then call getfile num
    end
    otherwise
   end
  end
  otherwise  /* the 'put' part is still missing */
 end
end

/* OBEX disconnect and exit the script */
call irda_write 0 O_CMD.Disconnect||'00 03'x
rc = left(irda_read_obex_packet(),1)
if rc <> O_HI.Sucess then say 'We have a problem (1), rc= ' c2x(rc)
exit

/* ------------------------------ */
/*    functions and procedures    */
/* ------------------------------ */

irda_open:

 rc = stream( port,'c','open' )
 if rc <> 'READY:' then do
   say "Can't open port rc= " rc stream( port, description)
   exit -1
  end
 rc = stream( port,'c','close' )
 rc = RxAsyncOpen( port, 'PortHandle' )
 if rc <> 0 then do
   say "Can't RxAsyncOpen port rc= " rc
   exit -1
  end
 return

/* ------------------------------ */

irda_write: procedure expose PortHandle Remaining debug logfile crlf

 parse arg delay packet
 call hexlog packet '>>> out >>>'
 rc=RxAsyncWrite(PortHandle, delay, packet, 'Remaining')
 if rc <> 0
  then say 'Error writing chars to irda, >' d2x(Remaining)||"'x < chars unwritten"
  else return
 exit -2
 return

/* ------------------------------ */

irda_write_at: parse ARG delay packet /* very minimalistic (dumb) function     */
                                      /* in this form only useful for AT*EOBEX */
 answer=''
 echo_answer=''
 call hexlog packet '>>> out >>>'
 rc=RxAsyncWrite( PortHandle, delay, packet||'0D'x, 'Remaining' )
 if rc <> 0
  then say 'Error writing chars to irda, >'Remaining'< chars unwritten'
  else do
   echo_answer=irda_read_line()
   call hexlog echo_answer '<<< in <<<<'
   do until answer='CONNECT'||crlf
    answer=irda_read_line()
    call hexlog answer '<<< in <<<<'
   end /*do*/
   return
  end /*do*/
 exit -2

/* ------------------------------ */

irda_read_line:  /* not used at the moment, useful if you alter the previous function */

 InpStr=''
 /* Syntax:    rc = RxAsyncRead( port, length, timeout, 'varname' )   */
 rc=RxAsyncRead( PortHandle, 0, 50, 'InpStr' )
 call testrcRxAR
 call hexlog InpStr '<<< in <<<<'
 return InpStr

/* ------------------------------ */

irda_read_obex_packet:  

 obex_packet=''
 InpStr=''
 /* Syntax:    rc = RxAsyncRead( port, length, timeout, 'varname' )   */
 rc=RxAsyncRead( PortHandle, 3, 10, 'InpStr' )
 call testrcRxAR
 obex_packet=InpStr
 oplen=c2d(right(InpStr,2))-3
 do while oplen > 0
  if oplen > 250 then do                              /* read in chunks of 255 bytes */
    rc=RxAsyncRead( PortHandle, 250, 200, 'InpStr' )  /* restriction of RxAsyncRead  */
    call testrcRxAR
    oplen=oplen-250
   end
  else do
    rc=RxAsyncRead( PortHandle, oplen, 3, 'InpStr' )
    call testrcRxAR
    oplen=0
   end 
  obex_packet=obex_packet||InpStr
 end
 call hexlog obex_packet '<<< in <<<<'
 return obex_packet

/* ------------------------------ */

testrcRxAR:  /* check for RxAsyncRead erors */

 if rc = 0 then return
  else if rc = -1 then say 'RxAsyncRead timed out, returned >' c2x(InpStr) '<'
  else say 'RxAsyncRead failed with rc=>'rc'<'
  exit -3
 return

/* ------------------------------ */

listcontent: /* list the folders first and then list the files */

 call irda_write 0 O_GetFolderListing_Packet
 buffer=irda_read_obex_packet()
 call charout , col.numatrib||format(0,3)||col.defatrib col.foldatrib||'RootDir'||col.defatrib||' [TopDir]'||listsep
 if listsep <> crlf then say
 call listfolders buffer
 call listfiles buffer
 say
 return

/* ------------------------------ */

/* list folders converted from UTF-8 to current codepage   */
/* just assuming that encoding from moblie is always UTF-8 */
listfolders: procedure expose col. folders. logfile listsep crlf debug columns cp

 parse arg packet
 needle_b='<folder name="'
 folders.0=countstr(needle_b,packet)
 do i= 1 to folders.0
  p1=pos(needle_b, packet)
  packet=substr(packet,p1+length(needle_b))
  needle_e='"/>'
  p2=pos(needle_e, packet)
  folders.i=left(packet,p2-1)
  call charout , col.numatrib||format(i,3)||col.defatrib col.foldatrib||UniXlat(folders.i,cp,'UTF-8')||col.defatrib||' [SubDir]'||listsep
  next=i+1
  parse value SysCurPos() with row column
  if (column + length(folders.next) + 16) > columns & listsep <> crlf & i < folders.0 then call charout , crlf
  if debug > 4 then  call hexlog u8t16(folders.i) '--- u8t16 ---'
 end
 return

/* ------------------------------ */

/* list files converted from UTF-8 to current codepage     */
/* just assuming that encoding from moblie is always UTF-8 */
/* example response to parse from my SE K310i: <file name="DSC00007.JPG" size="37603"/> */
listfiles: procedure expose col. files. logfile listsep crlf debug columns cp

 parse arg packet
 needle_1='<file name="'
 files.name.0=countstr(needle_1,packet)
 if files.name.0 > 0 then do
   do i= 1 to files.name.0
    p1=pos(needle_1, packet)
    packet=substr(packet,p1+length(needle_1))
    needle_2='" size="'          /* needle end */
    p2=pos(needle_2, packet)
    files.name.i=left(packet,p2-1)
    packet=substr(packet,p2+length(needle_2))
    needle_3='"/>'
    p3=pos(needle_3, packet)
    files.size.i=left(packet,p3-1)
    call charout , col.numatrib||format(i,3)||col.defatrib col.fileatrib||UniXlat(files.name.i,cp,'UTF-8')||col.defatrib format(files.size.i/1000,,1)||' kB'
    next=i+1
    parse value SysCurPos() with row column
    if (column + length(files.next) + 16) > columns & listsep <> crlf then call charout , crlf
    else call charout , listsep
    if debug > 4 then call hexlog u8t16(files.name.i) '--- u8t16 ---'
   end
 end
 return

/* ------------------------------ */

setpath: procedure expose folders. O_HI. O_CMD. buffer PortHandle Remaining cp debug crlf logfile O_SetPath_OH.

 parse arg folder
 if folder ='0' then O_SetPath_OH.1.content=''
 else O_SetPath_OH.1.content=u8t16(folders.folder)||'00 00'x
 O_SetPath_OH.1.len=length(O_SetPath_OH.1.HI||"00"||O_SetPath_OH.1.content)
 O_SetPathTotalLen=5+O_SetPath_OH.1.len
 O_SetPath_Complete=O_CMD.SetPath||x2c(d2x(O_SetPathTotalLen,4))||O_SetPath_OH.1.flags||O_SetPath_OH.1.constants||O_SetPath_OH.1.HI||x2c(d2x(O_SetPath_OH.1.len,4))||O_SetPath_OH.1.content
 call irda_write 0 O_SetPath_Complete
 rc=left(irda_read_obex_packet(),1)
 if rc <> O_HI.Sucess then say 'We have a problem (3), rc= ' c2x(rc)
 return

/* ------------------------------ */

delfile: procedure expose files. O_HI. O_CMD. buffer PortHandle Remaining cp debug crlf logfile

 parse arg file
 O_Put_OH.1.HI=O_HI.Name
 O_Put_OH.1.content=u8t16(files.name.file)||'00 00'x
 O_Put_OH.1.len=length(O_Put_OH.1.HI||"00"||O_Put_OH.1.content)
 O_PutTotalLen=3+O_Put_OH.1.len
 O_Put_Complete=O_CMD.Put||x2c(d2x(O_PutTotalLen,4))||O_Put_OH.1.HI||x2c(d2x(O_Put_OH.1.len,4))||O_Put_OH.1.content
 call irda_write 0 O_Put_Complete
 buffer=irda_read_obex_packet()
 rc=left(buffer,1)
 if rc<> O_HI.Sucess then  say 'We could not delete  HI_3=' c2x(rc)
 return

/* ------------------------------ */

getfile: procedure expose files. O_HI. O_CMD. buffer PortHandle Remaining cp debug CRLF logfile

 parse arg file
 O_Get_OH.1.HI=O_HI.Name
 O_Get_OH.1.content=u8t16(files.name.file)||'00 00'x
 O_Get_OH.1.len=length(O_Get_OH.1.HI||"00"||O_Get_OH.1.content)
 O_GetTotalLen=3+O_Get_OH.1.len
 O_Get_Complete=O_CMD.Get||x2c(d2x(O_GetTotalLen,4))||O_Get_OH.1.HI||x2c(d2x(O_Get_OH.1.len,4))||O_Get_OH.1.content
 call irda_write 0 O_Get_Complete
 buffer= irda_read_obex_packet()
 pac.HI.1=left(buffer,1)
 if pac.HI.1 <> O_HI.Cont & pac.HI.1 <> O_HI.Sucess then do
   say 'We have a problem (4) HI_1=' c2x(left(buffer,1)) ; exit
  end
 else do
  pac.len=c2d(substr(buffer,2,2))
  pac.con_id=c2d(substr(buffer,5,4))
  if substr(buffer,9,1) <> O_HI.Name then do
    say 'We have a problem (5)  HI_3=' c2x(substr(buffer,9,1)) ;  exit
   end
  else do
   pac.name.namelen=c2d(substr(buffer,10,2))
   pac.name.name=u16tcp(substr(buffer,12,pac.name.namelen-5))
   pac.data.len=c2d(substr(buffer,8+pac.name.namelen+2,4))
   pac.data.body.type=substr(buffer,8+pac.name.namelen+6,1)
   pac.data.body.len=c2d(substr(buffer,8+pac.name.namelen+7,2))
   pac.data.data=substr(buffer,8+pac.name.namelen+9)
  end
 end
 if stream(pac.name.name, 'c', 'query exists') <> '' then do
  say 'File exists! Overwrite? [Y/N]'
  pull ans
  if ans <> 'Y' then do
   call irda_write 0 O_CMD.Abort||'00 03'x
   say 'response from abort:' c2x(irda_read_obex_packet())
   end
 end
 if stream( pac.name.name, 'c', 'open write replace') <> 'READY:' then do
   say "We have a problem (6) can't open file for writing" ; exit
  end
 else call charout pac.name.name, pac.data.data
 if pac.HI.1 = O_HI.Sucess then do /* small file only one chunk was necessary */
   call stream pac.name.name, 'c', 'close' ; return
  end
 else do until pac.HI.1 = O_HI.Sucess | pac.HI.1 <> O_HI.Cont
   call irda_write 0 O_CMD.Get||'00 03'x /* get next chunk */
   buffer=irda_read_obex_packet()
   pac.HI.1=left(buffer,1)
   if pac.HI.1 <> O_HI.Cont & pac.HI.1 <> O_HI.Sucess then do
     say 'We have a problem (7) HI_1=' c2x(pac.HI.1)
    end
   else do
    pac.len=c2d(substr(buffer,2,2))
    pac.con_id=c2d(substr(buffer,5,4))
/*   if pac.con_id<>connection_id then iterate /* not our packet ? */ */
    pac.data.body.type=substr(buffer,9,1)
    pac.data.body.len=c2d(substr(buffer,10,2))
    pac.data.data=substr(buffer,12)
    call charout pac.name.name, pac.data.data
   end
  end
 call stream  pac.name.name, 'c', 'close'
 if pac.HI.1 <> O_HI.Sucess then say 'We had a problem (7) HI_1=' c2x(pac.HI.1)
 return

/* ------------------------------ */

u8t16: procedure /* convert and change byte order from intel to 'network' */
 parse arg buf
 temp1=UniXlat(buf,'UNICODE','UTF-8')
 temp2=''
 do i= 1 to length(temp1) by 2
  temp2=temp2||reverse(substr(temp1,i,2))
 end
 return temp2

/* ------------------------------ */

u16tcp: procedure expose cp debug crlf /* change byte order from 'network to intel */
 parse arg buf                         /* and convert to local codepage            */
 temp=''
 do i= 1 to length(buf) by 2
  temp=temp||reverse(substr(buf,i,2))
 end
 return UniXlat(temp,cp,'UNICODE')     /* errorhandling is missing */

/* ------------------------------ */

hexlog: procedure expose logfile crlf debug  /* some formatting for logfile output */
 parse arg string_log description
 if debug > 3 then do
  call charout logfile, description||crlf
  hexv=c2x(string_log)
  do i= 2 to length(string_log)*3 by 3
   hexv=insert(" ",hexv,i)
  end
  call charout logfile, hexv||'('||length(string_log)||') ('||d2x(length(string_log))||' Hex)'crlf
  if debug > 4 then call charout logfile, string_log||crlf
  call charout logfile, description||crlf
 end
 return

/* ------------------------------ */
